"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""
import plotly.graph_objs as go
import numpy as np
from util import arg_to_latex, beautify_latex, cast_0_dim
from plotly import figure_factory as ff
import pandas as pd
from collections import defaultdict


def line_plot(result, titles, verbose = False, **kwargs):
	'''N-d line plot f(t)'''
	f = result[titles['variable']]
	t_name, t = list(result.items())[0]
	if len(result) == 2:
		if len(f.shape) == 1:
			if verbose:
				print('\t1-d output', f.shape)
			assert f.shape == t.shape
			y = f
			layout = go.Layout(
				title = titles['title'], 
				xaxis = dict(title = '${}$'.format(t_name)),
				yaxis = dict(title = titles['title_lhs']))
			trace = go.Scatter(x = t, y = y, name = titles['title_short'])
			chart_type = 'line'

		elif f.shape[1] == 2:
			if verbose:
				print('\t2-d output', f.shape)
			x = f[:,0]
			y = f[:,1]
			# t_name, t = result.items()[0]
			text = ["{}:{}".format(t_name, v) for v in t]
			layout = go.Layout(title = titles['title'], 
							   xaxis = dict(title = 'x' + titles['units']),
							  yaxis = dict(title = 'y' + titles['units']))
			trace = go.Scatter(x = x, y = y, text = text)
			chart_type = '2d-line'

		elif f.shape[1] == 3: 
			if verbose:
				print('\t3-d output', f.shape)
			x = f[:,0]
			y = f[:,1]
			z = f[:,2]
			# t_name, t = result.items()[0]
			text = ["{}:{}".format(t_name, v) for v in t]
			layout = go.Layout(
				title = titles['title'],
				scene = dict(
					xaxis = dict(title = 'x' + titles['units']),
					yaxis = dict(title = 'y' + titles['units']),
					zaxis = dict(title = 'z' + titles['units']),
					))
			trace = go.Scatter3d(x = x, y = y, z = z, text = text, mode = 'lines')
			chart_type = '3d-line'

	elif len(result) == 4:
		arg0, val0 = t_name, t
		arg1, val1 = list(result.items())[1]
		arg2, val2 = list(result.items())[2]

		assert(f.shape == t.shape)

		text = ["{}:{}".format(titles['variable'], v) for v in f]
		
		trace = go.Scatter3d(
			x = val0, 
			y = val1, 
			z = val2, 
			text = text,
			marker = dict(
				color = result[titles['variable']], 
				colorscale = 'Viridis',
				colorbar = dict(title = titles['title_short'])),
			line = dict(
				color = result[titles['variable']],
				colorscale='Viridis'), 
			mode = 'markers+lines')
		layout = go.Layout(
			title = titles['title'],
			scene = dict(
				xaxis = dict(title = arg0),
				yaxis = dict(title = arg1),
				zaxis = dict(title = arg2),))
		chart_type = '4d-line'
	else:
		raise NotImplementedError('shape not supported: {}'.format(f.shape))

	return [trace], chart_type, layout


def vector_plot(result, titles, verbose = False, **kwargs):
	variable = titles['variable']
	val0 = list(result.values())[0]

	if (result[variable].shape == val0.shape) & (val0.shape[1] == 2):
		if verbose:
			print('\t 2-d output', result[variable].shape)
		u = result[variable][:,0]
		v = result[variable][:,1]
		x = val0[:,0]
		y = val0[:,1]

		trace = ff.create_quiver(
			x, y, u, v,
			# scale=.25,
			# arrow_scale=.4,
			name='quiver',
			line=dict(width=1))['data'][0] # extracts quiver trace

		layout = go.Layout(title = titles['title'], 
							  xaxis = dict(title = 'x'),
							  yaxis = dict(title = 'y'))
		chart_type = '2d-vector'

	elif (result[variable].shape == val0.shape) & (val0.shape[1] == 3):
		if verbose:
			print('\t 3-d output', result[variable].shape)
		u = result[variable][:,0].tolist()
		v = result[variable][:,1].tolist()
		w = result[variable][:,2].tolist()
		x = val0[:,0].tolist()
		y = val0[:,1].tolist()
		z = val0[:,2].tolist()
		trace = go.Cone(x = x, y = y, z = z, u = u, v = v, w = w)

		layout = go.Layout(
			title = titles['title'],
			scene = dict(
				xaxis = dict(title = 'x'),
				yaxis = dict(title = 'y'),
				zaxis = dict(title = 'z'),
				))
		chart_type = '3d-vector'
	else:
		raise NotImplementedError('plot type not supported yet')

	return [trace], chart_type, layout

def contour_plot(result, titles, indexing, verbose = False, **kwargs):
	variable = titles['variable']
	if verbose:
		print('\t2-d output', result[variable].shape)
	z = result[variable]
	title = titles['title']
	arg0, val0 = list(result.items())[0]
	arg1, val1 = list(result.items())[1]

	if verbose:
		print('2 inputs: {} {}, {} {}'.format(arg0, val0.shape, arg1, val1.shape))
	traces = []
	if len(val0.shape) == 1:
		if verbose:
			print('\t\t1-d args', end=' ')
		if result[variable].shape[0] == val0.shape[0]:
			if indexing == 'ij':
				if verbose:
					print('{} indexing'.format(indexing))
				trace = go.Contour(x = val0,
								   y = val1,
								   z = z.T)
			else:
				if verbose:
					print('{} indexing'.format(indexing))
				trace = go.Contour(x = val0,
								   y = val1,
								   z = z)

		else:
			if verbose:
				print('xy indexing')
			trace = go.Contour(x = val0,
							   y = val1,
							   z = z)
		layout = go.Layout(
			title = title,
			xaxis = dict(title = '${}$'.format(arg0)),
			yaxis = dict(title = '${}$'.format(arg1)))
		traces.append(trace)
		chart_type = '2d-grid'
	else:
		if verbose:
			print('\t\t2-d args', val0.shape, val1.shape)
		assert val0.shape == val1.shape
		assert val0.shape == z.shape
		xaxis = dict(title = beautify_latex('${}$'.format(arg_to_latex(arg0))))
		yaxis = dict(title = beautify_latex('${}$'.format(arg_to_latex(arg1))))
		carpet_traces, layout = carpet_plot(result, title, xaxis, yaxis)
		traces.extend(carpet_traces)
		chart_type = '2d-skew'

	return traces, chart_type, layout

def carpet_plot(results, title, xaxis, yaxis, indexing = 'xy', **kwargs):
	'''Assumes ordered dict where values have the same shape'''
	# print title, xaxis, yaxis
	arg0, val0 = list(results.items())[0]
	arg1, val1 = list(results.items())[1]
	arg2, val2 = list(results.items())[2]
	
	a = list(range(val0.shape[0]))
	b = list(range(val0.shape[1]))

	aa, bb = np.meshgrid(b,a, indexing = indexing)

	trace1 = go.Contourcarpet(
		a = aa.ravel(),
		b = bb.ravel(),
		z = val2.ravel(),
	)

	trace2 = go.Carpet(
		a = aa.ravel(),
		b = bb.ravel(),
		x = val0.ravel(),
		y = val1.ravel(),
	)
	traces = [trace1, trace2]
	layout = go.Layout(
		title = title, 
		xaxis = xaxis,
		yaxis =  yaxis)
	return traces, layout

def plane(result, titles, indexing = 'xy', verbose = False, **kwargs):
	variable = titles['variable']
	arg0, val0 = list(result.items())[0]
	arg1, val1 = list(result.items())[1]
	arg2, val2 = list(result.items())[2]
	traces = []
	if verbose:
		print('\t3-d output', result[variable].shape)
	surfacecolor = result[variable].squeeze()
	create_meshgrid = True
	for x in val0, val1, val2:
		if len(get_arg_shapes(x)[0]) != 1:
			create_meshgrid = False
			ones = np.ones_like(x)
	if create_meshgrid:
		xx, yy, zz = np.meshgrid(val0, val1, val2)
	else:
		xx, yy, zz = [v*ones for v in [val0, val1, val2]]
	traces.append(go.Surface(
		x = xx.squeeze(),
		y = yy.squeeze(),
		z = zz.squeeze(),
		surfacecolor = surfacecolor))
	layout = go.Layout(
		title = titles['title'],
		scene = dict(
			xaxis = dict(title = arg0),
			yaxis = dict(title = arg1),
			zaxis = dict(title = arg2),
		)
	)
	chart_type = '3d-plane'

	return traces, chart_type, layout

def surface(result, titles, verbose = False, **kwargs):
	variable = titles['variable']
	title = titles['title']
	arg0, val0 = list(result.items())[0]
	arg1, val1 = list(result.items())[1]
	arg2, val2 = list(result.items())[2]

	if verbose:
		print('\t2-d output', result[variable].shape)

	traces = []

	if len(result[variable].shape) == 2:
		surfacecolor = result[variable]
		xx = cast_0_dim(val0, result[variable])
		yy = cast_0_dim(val1, result[variable])
		zz = cast_0_dim(val2, result[variable])
		traces.append(go.Surface(
			x = xx,
			y = yy,
			z = zz,
			surfacecolor = surfacecolor))
		layout = go.Layout(
			title = title,
			scene = dict(
				xaxis = dict(title = arg0),
				yaxis = dict(title = arg1),
				zaxis = dict(title = arg2),))
		chart_type = '3d-surface-scalar'

	elif len(result[variable].shape) == 1:
		if verbose:
			print('\tscalar output', result[variable].shape)

		elif result[variable].shape[0] == 1:
			traces.append(go.Surface(
				x = val0, 
				y = val1, 
				z = val2, 
				))
			layout = go.Layout(
				title = title,
				scene = dict(
					xaxis = dict(title = arg0),
					yaxis = dict(title = arg1),
					zaxis = dict(title = arg2),
					))
			chart_type = '3d-surface'

	elif len(result[variable]) == 1: #scalar or color value?
		if verbose:
			print('\tsingle valued output', result[variable].shape)
		surfacecolor = result[variable]
		arg0, val0 = list(result.items())[0]
		arg1, val1 = list(result.items())[1]
		arg2, val2 = list(result.items())[2]
		xx = val0
		yy = val1
		zz = val2
		traces.append(go.Surface(
			x = xx,
			y = yy,
			z = zz,
			))
		layout = go.Layout(
			title = title,
			scene = dict(
				xaxis = dict(title = arg0),
				yaxis = dict(title = arg1),
				zaxis = dict(title = arg2),
			)
		)

	return traces, chart_type, layout



plot_dict = {
	(1,)	:	{(('N','M'), ('N','M'), ('N','M')): {'name': '3d-parametric', 'func': surface}},
	('N',)	:	{
		(('N',),) : {'name': '1d-line', 'func': line_plot},
		(('N',),('N',),('N',)): {'name': '3d-line-scalar', 'func': line_plot},},
	('N',2) :	{
		(('N',),) :{'name': '2d-line', 'func': line_plot},
		(('N',2),):{'name': '2d-vector', 'func': vector_plot},
	},
	('N',3) : 	{
		(('N',),) :{'name': '3d-line', 'func': line_plot},
		(('N',3),):{'name': '3d-vector', 'func': vector_plot},
	},
	('N','M'):	{
		(('N',),('M',)) :{'name': '2d-contour', 'func': contour_plot},
		(('N','M'),('N','M')):{'name': '2d-contour-skew', 'func': contour_plot},
		(('N','M'), ('N','M'), ('N','M')): {'name': '3d-parametric-scalar', 'func': surface},
		((1,),('N','M'),('N','M')):{'name': '3d-plane', 'func': plane},
		(('N','M'),(1,),('N','M')):{'name': '3d-plane', 'func': plane},
		(('N','M'),('N','M'),(1,)):{'name': '3d-plane', 'func': plane},
	},
	('N','M',1): {
		((1,),('N',),('M',)):{'name': '3d-plane', 'func': plane},
		(('N',),(1,),('M',)):{'name': '3d-plane', 'func': plane},
		(('N',),('M',),(1,)):{'name': '3d-plane', 'func': plane},
	},
}

def get_arg_shapes(*args):
	shapes = []
	for a in args:
		if type(a) == np.ndarray:
			shape = a.shape
		else:
			try:
				shape = (len(a),)
			except:
				shape = (1,)
		shapes.append(shape)
	return shapes


def get_plot_key(out_shape, *arg_shapes):
	shapes_match = all([(a == out_shape) for a in list(arg_shapes)])
	nargs = len(arg_shapes)
	arg_dims = ''
	if nargs == 1:
		if len(out_shape) == 1:
			out_dim = 'N'
		else:
			out_dim = ('N', out_shape[-1])
	else:
		if len(out_shape) == 1:
			if out_shape[0] == 1:
				out_dim = 1,
			else:
				out_dim = 'N',
		elif len(out_shape) == 2:
			out_dim = 'N','M'
		elif len(out_shape) == 3:
			if 1 in out_shape:
				out_dim = 'N', 'M', 1
			else:
				out_dim = 'N','M','L'
	
	out_dim = tuple(out_dim)
	if shapes_match:
		arg_dims = tuple(nargs*[out_dim])
	else:
		if nargs == 1:
			if out_dim == ('N',2):
				if arg_shapes[0][0] == out_shape[0]:
					arg_dims = (('N',),)
			elif out_dim == ('N',3):
				if arg_shapes[0][0] == out_shape[0]:
					arg_dims = (('N',),)
		elif nargs == 2:
			if out_dim == ('N','M'):
				if (arg_shapes[0][0] in out_shape) & (arg_shapes[1][0] in out_shape):
					arg_dims = (('N',),('M',))
		elif nargs == 3:
			if out_dim == (1,):
				if len(set(arg_shapes)) == 1:
					if len(arg_shapes[0]) == 2:
						arg_dims = tuple(3*[('N','M')])
			elif out_dim == ('N','M'):
				arg_set = set([1])
				for arg in arg_shapes:
					arg_set.update(set(arg))
				if arg_set - set(out_shape) == set([1]):
					if arg_shapes[0] == (1,):
						arg_dims = ((1,), ('N','M'), ('N','M'))
					elif arg_shapes[1] == (1,):
						arg_dims = (('N','M'),(1,),('N','M',))
					elif arg_shapes[2] == (1,):
						arg_dims = (('N','M'),('N','M'),(1,))
			elif out_dim == ('N','M',1):
				arg_set = set([1])
				for arg in arg_shapes:
					arg_set.update(set(arg))
				if arg_shapes[0] == (1,):
					arg_dims = ((1,), ('N',), ('M',))
				elif arg_shapes[1] == (1,):
					arg_dims = (('N',),(1,),('M',))
				elif arg_shapes[2] == (1,):
					arg_dims = (('N',),('M',),(1,))
	if arg_dims == '':
		raise NotImplementedError('No way to handle out_shape {}, with arg shapes:{}'.format(out_shape, arg_shapes))
	return out_dim, arg_dims


plot_types = dict()
for out_shape, v in list(plot_dict.items()):
	for arg_shapes, v_ in list(v.items()):
		plot_key = get_plot_key(out_shape,*arg_shapes)
		if plot_key in plot_types:
			raise KeyError('plot_key already present {}'.format(plot_key))
		plot_types[plot_key] = [v_['name'], v_['func']]

plot_types = pd.DataFrame(plot_types).T
plot_types.index.set_names(['out_shape', 'arg_shapes'], inplace = True)
plot_types.columns = ['plot_type', 'function']

def test_plot_keys():
	for k in plot_types.to_dict(orient = 'index'):
		try:
			plot_dict[k[0]][k[1]]['name']
		except KeyError:
			print('could not find', k[0], k[1])
			raise




"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""
import os
import tempfile
import sys
import numpy.f2py  # just to check it presents
from numpy.distutils.exec_command import exec_command
from collections import OrderedDict, Callable, defaultdict
import functools
from sympy import sympify
# from pint import UnitRegistry
from sympy.utilities.autowrap import ufuncify
import functools
from decorator import decorator, decorate
from sympy import symbols, Symbol
from sympy.core.function import UndefinedFunction
from inspect import getargspec
from sympy.physics import units
from sympy.physics import units as sympy_units
import numpy as np
from sympy import latex, Eq
from sympy.parsing.latex import parse_latex
import pandas as pd



def get_unit_quantity(name, base, scale_factor, abbrev = None, unit_system = 'SI'):
	'''Define a unit in terms of a base unit'''
	u = units.Quantity(name, abbrev = abbrev)
	base_unit = getattr(sympy_units, base)
	u.set_dimension(base_unit.dimension)
	u.set_scale_factor(scale_factor*base_unit, unit_system = unit_system)
	return u

unit_subs = dict(nT = get_unit_quantity('nanotesla', 'tesla', .000000001, 'nT', 'SI'),
				R_E = get_unit_quantity('earth radii', 'km', 6.371e6, 'R_E', 'SI'),
				erg = get_unit_quantity('erg', 'J', .0000001, 'erg', 'SI'),)

# global_ureg = UnitRegistry()
# global_ureg.define('m^3 = m * m * m = m3')
# global_ureg.define('cc = cm * cm * cm')
# global_ureg.define('R_s = 6.957e8 m')
# global_ureg.define('AU = 1.496e+11 m')

def get_ufunc(expr, variable_map):    
	expr = sympify(expr, locals = variable_map)
	f = ufuncify(expr.free_symbols, expr)
	formula = 'f{} = {}'.format(tuple(expr.free_symbols), expr)
	return f, formula

def Quantity_Factory(ureg = None):
	"""Emits Quantity class from unit registry that supports unit bracket syntax"""
	if ureg is None:
		ureg = global_ureg

	class Quantity(ureg.Quantity):
		def __init__(self, *args, **kwargs):
			super(Quantity, self).__init__(*args, **kwargs)

		def __getitem__(self, key):
			"""Extends [] notation to support unit conversion"""
			try:
				super(Quantity, self).__getitem__(key)
			except:
				return self.to(key)

	return Quantity


def Quantity(fun, ureg = None):
	"""Wraps a function as a Quantity"""
	if ureg is None:
		ureg = global_ureg
	def wrapper(*args, **kwargs):
		return Quantity_Factory(ureg)(fun(*args, **kwargs), fun.units)
	return wrapper

def compile_fortran(source, module_name, extra_args='', folder = './'):
	with tempfile.NamedTemporaryFile('w', suffix='.f90') as f:
		f.write(source)
		f.flush()

		args = ' -c -m {} {} {}'.format(module_name, f.name, extra_args)
		command = 'cd "{}" && "{}" -c "import numpy.f2py as f2py;f2py.main()" {}'.format(folder, sys.executable, args)
		status, output = exec_command(command)
		return status, output, command

def substr_replace(name, name_maps):
	"""replaces all substrings in name with those given by name_maps"""
	for old, new in name_maps:
		name = name.replace(old, new)
	return name

def beautify_latex(s):
	return substr_replace(s, [
		('plus', '+'),
		('minus', '-'),
		('comma', ','),
		('LEFT', '\\left ('),
		('RIGHT', '\\right )'),
		('integral', '\\int'),
		])
	



def arg_to_latex(arg):
	return beautify_latex(latex(Symbol(arg)))

"""Overloading technique found here
https://stackoverflow.com/a/25344433/2793333"""
def multidispatch(*types, **kwargs):
	def register(function):
		name = function.__name__
		mm = multidispatch.registry.get(name)
		if mm is None:
			@functools.wraps(function)
			def wrapper(self, *args):
				types = tuple(arg.__class__ for arg in args) 
				function = wrapper.typemap.get(types)
				if function is None:
					raise TypeError("no match")
				return function(self, *args)
			wrapper.typemap = {}
			mm = multidispatch.registry[name] = wrapper
		if types in mm.typemap:
			raise TypeError("duplicate registration")
		if kwargs.get('verbose', False):
			print('registering types', str(tuple([t.__name__ for t in types])))
		mm.typemap[types] = function
		return mm
	return register
multidispatch.registry = {}




class DefaultOrderedDict(OrderedDict):
	# Source: http://stackoverflow.com/a/6190500/562769
	def __init__(self, default_factory=None, *a, **kw):
		if (default_factory is not None and
		   not isinstance(default_factory, Callable)):
			raise TypeError('first argument must be callable')
		OrderedDict.__init__(self, *a, **kw)
		self.default_factory = default_factory

	def __getitem__(self, key):
		try:
			return OrderedDict.__getitem__(self, key)
		except KeyError:
			return self.__missing__(key)

	def __missing__(self, key):
		if self.default_factory is None:
			raise KeyError(key)
		self[key] = value = self.default_factory()
		return value

	def __reduce__(self):
		if self.default_factory is None:
			args = tuple()
		else:
			args = self.default_factory,
		return type(self), args, None, None, list(self.items())

	def copy(self):
		return self.__copy__()

	def __copy__(self):
		return type(self)(self.default_factory, self)

	def __deepcopy__(self, memo):
		import copy
		return type(self)(self.default_factory,
						  copy.deepcopy(list(self.items())))

	def __repr__(self):
		return 'OrderedDefaultDict(%s, %s)' % (self.default_factory,
											   OrderedDict.__repr__(self))


def komodo_wrapper(f, *args, **kwargs):
	"""Wrapper needed by decorator.decorate to pass through args, kwargs"""
	return f(*args, **kwargs)


def kamodofy(_func = None, units = '', data = None, update = None, equation = None, citation = None, hidden_args = [], **kwargs):
	"""Adds meta and data attributes to functions for compatibility with Komodo

	meta: a dictionary containing {units: <str>}
	data: 
		if supplied, set f.data = data
		if not supplied, set f.data = f(), assuming it can be called with no arguments.
			If f cannot be called with no arguments, set f.data = None
	"""
	def decorator_kamodofy(f):
		f.meta = dict(units = units, citation = citation, equation = equation, hidden_args = hidden_args)
		if citation is not None:
			f.__doc__ = f.__doc__ + '\n\ncitation: {}'.format(citation)
		f.update = update
		if data is None:
			try:
				f.data = f(**kwargs)
			except:
				f.data = None
		else:
			f.data = data

		if equation is not None:
			latex_str = equation.strip("$")
			f._repr_latex_ = lambda : latex_str
			# f._repr_latex_ = lambda : "${}$".format(latex(parse_latex(latex_str)))
		else:
			f_ = symbols(f.__name__, cls = UndefinedFunction)
			lhs = f_.__call__(*symbols(getargspec(f).args))
			lambda_ = symbols('lambda', cls = UndefinedFunction)
			latex_eq = latex(Eq(lhs, lambda_(*lhs.args)))
			f._repr_latex_ = lambda : "${}$".format(latex(latex_eq))




		return decorate(f,komodo_wrapper) #preserves signature
	
	if _func is None:
		return decorator_kamodofy
	else:
		return decorator_kamodofy(_func)

def symbolic(sym_names):
	"""Returns symbolic functions""" 
	syms = symbols(sym_names, cls = UndefinedFunction)
	if len(syms) == 1:
		return syms[0]
	else:
		return syms


def sort_symbols(symbols):
	symbols_ = list(symbols)
	symbols_.sort(key = str)
	return tuple(symbols_)


def valid_args(f, kwargs):
	'''Extract arguments from kwargs that appear in f'''
	valid = OrderedDict()
	if type(f) == np.vectorize:
		for a in getargspec(f.pyfunc).args:
			if a in kwargs:
				valid[a] = kwargs[a]
		return valid
	else:		
		for a in getargspec(f).args:
			if a in kwargs:
				valid[a] = kwargs[a]
		return valid

def eval_func(func, kwargs):
	'''Evaluate function over valid argument'''
	try:
		return func(**valid_args(func, kwargs))
	except TypeError as m:
		raise TypeError(str(m) + str(list(valid_args(func, kwargs).keys())))

def get_defaults(func):
	spec = getargspec(func)
	args = spec.args
	defaults = spec.defaults
	if defaults is not None:
		return OrderedDict(list(zip(args[-len(defaults):], spec.defaults)))
	else:
		return None


def to_arrays(d):
	for k,v in list(d.items()):
		d[k] = np.asarray(v)
	return d

def cast_0_dim(a, to):
	if a.ndim == 0:
		return a*np.ones(to.shape)
	else:
		return a

def simulate(state_funcs, **kwargs):
	"""Iterate over functions. 
	
	state_funcs(OrderedDict)
		key: the variable to update
		value: the function that updates the variable
	
	Any remaining kwargs are passed to the state functions.
	
	The order of the keys in state_funcs determines which variables get updated first.
	"""
	
	verbose = kwargs.get('verbose', False)
	steps = kwargs.get('steps', 1)

	state_dict = kwargs.copy()
	yield OrderedDict([(k, state_dict.get(k,None)) for k in state_funcs])
	for i in range(steps):
		result = []
		for arg, f in list(state_funcs.items()):
			try:
				result.append((arg, eval_func(f, state_dict)))
			except TypeError as m:
				raise TypeError('{}:'.format(arg) + str(m))
			state_dict.update(OrderedDict(result))
		yield OrderedDict(result)

def pad_nan(array):
	if len(array.shape) == 1:
		return np.pad(array.astype(float), (0,1), 
					  mode = 'constant', constant_values = np.nan)
	elif len(array.shape) == 2:
		return np.pad(array.astype(float), ((0,1),(0,0)), 
					  mode = 'constant', constant_values = np.nan)
	else:
		raise NotImplementedError('cannot pad shape {}'.format(array.shape))

def concat_solution(gen, variable):
	result = []
	params = defaultdict(list)
	for f in gen:
		params[variable].append(pad_nan(f.data))
		for k,v in list(get_defaults(f).items()):
			params[k].append(pad_nan(v))
		
	for k,v in list(params.items()):
		params[k] = np.vstack(v)
	return params

	
existing_plot_types = pd.DataFrame({
	('1', 'N', 'N'): ['1-d line', ''],
	('1', 'N', 'Nx2'): ['2-d line',''],
	('1', 'N', 'Nx3'): ['3-d line', ''],
	('3', 'N, N, N', 'N'): ['3-d colored line', ''],
	( '1', 'Nx2', 'Nx2'): ['2-d vector field', ''],
	('1', 'Nx3', 'Nx3'): ['3-d vector field', ''],
	('2', 'N, M', 'NxM'): ['2-d contour', 'indexing'],
	('2', 'NxM, NxM', 'NxM'): ['2-d contour (skew/carpet)', 'indexing'],
	('3', 'NxM, NxM, NxM', '1'): ['Parametric Surface', ''],
	('3', 'NxM, NxM, NxM', 'NxM'): ['Coloured Parametric Surface', ''],
	('3', 'L, M, 1', 'LxM'): ['Map-to-plane', 'indexing*'],
	('3', '1, M, N', 'MxN'): ['Map-to-plane', 'indexing*'],
	('3', 'L, 1, N', 'LxN'): ['Map-to-plane', 'indexing*']
	}).T
existing_plot_types.index.set_names(['nargs', 'arg shapes', 'out shape'], inplace = True)
existing_plot_types.columns = ['Plot Type', 'notes']


def test_valid_args():
	def f(x,y,z):
		return x + y + z

	args = valid_args(f, dict(x = 1, y = 2, z = 3, g = 4))
	assert args['x'] == 1
	assert ('g' not in args)


def test_eval_func():
	def f(x, y):
		return x + y
	assert eval_func(f, dict(x = 1, y = 2, z = 3)) == 1 + 2


def test_simulate():
	def update_y(x):
		return x + 1

	def update_x(y):
		return y - 2

	state_funcs = OrderedDict([
			('y', update_y),
			('x', update_x),
			('t', lambda t, dt: t + dt)
		  ])
			
	simulation = simulate(state_funcs,
						  x = 3, #initial conditions
						  t = 0,
						  dt = 1,
						  steps = 10)

	for state in simulation:
		pass

	assert state['x'] == -7
	assert state['y'] == -5
	assert state['t'] == 10


def test_meta_units():
	@kamodofy(units = 'kg')
	def mass(x):
		return x

	assert mass.meta['units'] == 'kg'


def test_meta_data():
	@kamodofy(units = 'kg')
	def mass(x = list(range(5)), y = 3):
		return [x_ + y for x_ in x]

	assert mass.data[-1] == 7



def test_meta_data_kwargs():
	@kamodofy(x = 3, y = 3)
	def e(x,y):
		return x + y
	assert e.data == 6

	@kamodofy(x = 3)
	def f(x = 5, y = 3):
		return x + y
	assert f.data == 6

	@kamodofy()
	def g(x, y):
		return x + y
	assert g.data == None
	assert g.meta['units'] == ''

	@kamodofy(data = 3)
	def h(x,y):
		return x + y
	assert h.data == 3



def test_repr_latex():
	@kamodofy(equation = '$f(x) = x_i^2$')
	def f(x):
		return x

	assert f._repr_latex_() == '$f{\\left (x \\right )} = x_{i}^{2}$'


	@kamodofy
	def g(x,y,z):
		return x

	assert g._repr_latex_() == '$g{\\left (x,y,z \\right )} = \\lambda{\\left (x,y,z \\right )}$'

def test_bibtex():

	bibtex = """
@phdthesis{phdthesis,
  author       = {Peter Joslin}, 
  title        = {The title of the work},
  school       = {The school of the thesis},
  year         = 1993,
  address      = {The address of the publisher},
  month        = 7,
  note         = {An optional note}
}"""
	@kamodofy(citation = bibtex)
	def h(x):
		'''This equation came out of my own brain'''
		return x

	assert '@phdthesis' in h.meta['citation']





"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""
try:
	import pytest
except ImportError:
	pass

from collections import OrderedDict
import collections

import numpy as np
from sympy import Integral, Symbol, symbols, Function

from sympy.parsing.sympy_parser import parse_expr

from sympy import lambdify
from sympy.parsing.latex import parse_latex
from sympy import latex
from sympy.core.function import UndefinedFunction
from inspect import getargspec
from sympy import Eq
import pandas as pd
from IPython.display import Latex


from sympy.physics import units as sympy_units
from sympy.physics.units import Quantity
from sympy.physics.units import Dimension
from sympy import Expr
from sympy import sympify

import functools
from .util import kamodofy
from .util import sort_symbols
from .util import simulate
from .util import unit_subs
from .util import get_defaults, valid_args, eval_func
from .util import to_arrays, cast_0_dim
from .util import beautify_latex, arg_to_latex
from .util import concat_solution

import sympy.physics.units as u


import plotly.graph_objs as go
from plotly import figure_factory as ff

from plotting import *
from .util import existing_plot_types

from sympy import Wild
from types import GeneratorType

def get_unit_quantities():
	subs = {} 
	for k, v in list(sympy_units.__dict__.items()): 
		if isinstance(v, Expr) and v.has(sympy_units.Unit): 
			subs[Symbol(k)] = v # Map the `Symbol` for a unit to the quantity
	return subs


def clean_unit(unit_str):
	"""remove brackets and all white spaces in and around unit string"""
	return unit_str.strip().strip('[]').strip()

def get_unit(unit_str, unit_subs = unit_subs):
	unit_str = clean_unit(unit_str)
	units = get_unit_quantities()

	if unit_subs is not None:
		units.update(unit_subs)

	if len(unit_str) == 0:
		return Dimension(1)
	try:
		unit = parse_expr(unit_str.replace('^', '**')).subs(units)
	except:
		raise NameError('something wrong with unit str [{}], type {}'.format(unit_str, type(unit_str)))
	try:
		assert len(unit.free_symbols) == 0
	except:
		raise ValueError("Unsupported unit: {} {}".format(unit_str, type(unit_str)))
	return unit

def get_expr_with_units(expr, units_map):
	"""Replaces symbols with symbol*unit"""
	subs = []
	for s,u in list(units_map.items()):
		if type(s) == str:
			subs.append((s, Symbol(s)*u))
		else: #assume symbol
			subs.append((s,s*u))
	new_expr = expr.subs(subs)
	return new_expr

def get_expr_without_units(expr, to, units_map):
	"""Converts an expression with units to one without, applying conversion factors where necessary"""
	expr = sympy_units.convert_to(expr,to)
	subs = []
	for s in units_map:
		if type(s) == str:
			subs.append((Symbol(s)*to, Symbol(s)))
		else: #assume symbol
			subs.append((s*to, s))
	return expr.subs(subs)


def args_from_dict(expr, local_dict):
	args = []
	if local_dict is not None:
		for a in expr.args:
			if str(a) in local_dict:
				args.append(local_dict[str(a)])
			else:
				args.append(a)
		return tuple(args)
	else:
		return expr.args

def extract_units(lhs):
	if '[' in lhs:
		lhs, units = lhs.split('[')
		return lhs, units.split(']')[0]
	else:
		return lhs, ''

def expr_to_symbol(expr, args):
	"""f,args -> f(args) of class function, f(x) -> f(x)"""
	if type(expr) == Symbol:
		return parse_expr(str(expr) + str(args))
		# return symbols(str(expr), cls = Function)
	else:
		return expr 

def parse_lhs(lhs, local_dict):	
	lhs, units = extract_units(lhs)
	parsed = parse_expr(lhs)
	try:
		args = args_from_dict(parsed, local_dict)
	except:
		print(local_dict)
		raise
	symbol = expr_to_symbol(parsed, args)
	return symbol, args, units, parsed

def parse_rhs(rhs, is_latex, local_dict = None):
	parsed = dict()
	if is_latex:
		if local_dict is not None:
			expr = parse_latex(rhs).subs(local_dict)
		else:
			expr = parse_latex(rhs)
	else:
		expr = parse_expr(rhs).subs(local_dict)
	return expr

def parse_function(function, local_dict = None):
	parsed = dict()

	is_latex = False

	if "$" in function:
		is_latex = True
		function = function.strip("$").strip()
	
	split = [a.strip() for a in function.split('=')]
	
	if len(split) == 2:
		lhs, rhs = split
		expr = parse_rhs(rhs, is_latex, local_dict)
	else:
		lhs = split[0]
		expr = None

	symbol, args, units, lhs_expr = parse_lhs(lhs, local_dict)
	if expr is not None:
		if set(args) != expr.free_symbols:
			raise NameError("Mismatched symbols: {} and {}".format(args, expr.free_symbols))

	parsed['expr'] = expr
	parsed['symbol'] = symbol
	parsed['args'] = args
	parsed['units'] = units

	return parsed

def get_function_args(func, hidden_args = []):
	"""converts function arguments to list of symbols"""
	return symbols([a for a in getargspec(func).args if a not in hidden_args])

def wildcard(expr):
	result = expr
	for s in expr.free_symbols:
		result = result.replace(s, Wild(str(s)))
	return result


def validate_units(expr, units):
	result = expr.subs(units, simultaneous = True)
	if len(result.free_symbols) != 0:
		return Dimension(1)
	else:
		for s, unit in list(units.items()):
			try:
				result = result.replace(wildcard(s), unit)
			except:
				pass
		return result




def match_dimensionless_units(lhs_units, rhs_units):
	'''if lhs_units is dimensionless and rhs_units is not dimensionless, assign rhs_units to lhs_units (and vice versa)'''
	if lhs_units == Dimension(1): # f = ...
		if rhs_units != lhs_units: # f = rho[kg/m^3]
			lhs_units = rhs_units # f[kg/m^3] = rho[kg/m^3]
	elif rhs_units == Dimension(1): # ... = x
		if rhs_units != lhs_units: # f[kg/m^3] = x
			rhs_units = lhs_units # f[kg/m^3] = x[kg/m^3]
	return lhs_units, rhs_units
	

def check_unit_compatibility(rhs_units, lhs_units):
	try:
		assert sympy_units.convert_to(rhs_units + lhs_units, lhs_units)
	except:
		raise NameError('incompatible units:{} and {}'.format(lhs_units, rhs_units))


class Kamodo(collections.OrderedDict):
	'''Kamodo base class demonstrating common API for space weather models


	This API provides access to space weather fields and their properties through:

		interpolation of variables at user-defined points
		unit conversions
		coordinate transformations specific to space weather domains

	Required methods that have not been implemented in child classes will raise a NotImplementedError
	'''
	def __init__(self, *funcs, **kwargs):
		"""Base initialization method

		Args:
			param1 (str, optional): Filename of datafile to interpolate from

		"""

		super(Kamodo, self).__init__()
		self.symbol_registry = OrderedDict()

		symbol_dict = kwargs.pop('symbol_dict', None)

		self.verbose = kwargs.pop('verbose', False)
		self.signatures = OrderedDict()

		for func in funcs:
			if type(func) is str:
				lhs,rhs = func.split('=')
				self[lhs.strip('$')] = rhs

		for sym_name, expr in list(kwargs.items()):
			self[sym_name] = expr


	def __getitem__(self, key):
		try:
			return super(Kamodo, self).__getitem__(key)
		except:
			return self[self.symbol_registry[key]]

	def register_symbol(self, symbol):
		self.symbol_registry[str(type(symbol))] = symbol

	def load_symbol(self, sym_name):
		symbol = self.symbol_registry[sym_name]
		signature = self.signatures[str(symbol)]
		lhs_expr = signature['lhs']
		units = signature['units']
		return symbol, symbol.args, units, lhs_expr

	def remove_symbol(self, sym_name):
		if self.verbose:
			print('removing {} from symbol_registry'.format(sym_name))
		symbol = self.symbol_registry.pop(sym_name)
		if self.verbose:
			print('removing {} from signatures'.format(symbol))
		self.signatures.pop(str(symbol))

	def parse_key(self, sym_name):
		args = tuple()
		if sym_name not in self:
			if self.verbose:
				print('parsing new key {}'.format(sym_name))
			if type(sym_name) is str:
				sym_name = sym_name.strip('$').strip()
				if sym_name not in self.symbol_registry:
					symbol, args, units, lhs_expr = parse_lhs(sym_name, self.symbol_registry)
					if self.verbose:
						print('newly parsed symbol:', symbol, type(symbol))
					if str(type(symbol)) in self.symbol_registry:
						raise KeyError("{} found in symbol_registry".format(str(type(symbol))))
				else:
					raise KeyError("{} found in symbol_registry".format(sym_name))
			else:
				if type(sym_name) is Symbol:
					symbol = sym_name
					units = '' # where else do we get these?
					lhs_expr = symbol
				else:
					#cast the lhs into a string and parse it
					symbol, args, units, lhs_expr = parse_lhs(str(sym_name), self.symbol_registry) 
		else:
			symbol = sym_name
			if self.verbose:
				print('{} found in keys'.format(sym_name))
			try:
				units = self.signatures[sym_name]['units']
				lhs_expr = self.signaturs[sym_name]['lhs']
			except KeyError:
				units = ''
				lhs_expr = symbol
		return symbol, args, units, lhs_expr


	def parse_value(self, rhs_expr, local_dict = None):
		if type(rhs_expr) is str:
			is_latex = '$' in rhs_expr
			rhs_expr = rhs_expr.strip('$').strip()
			rhs_expr = parse_rhs(rhs_expr, is_latex, local_dict = local_dict)
		return rhs_expr


	def check_or_replace_symbol(self, symbol, free_symbols, rhs_expr = None):
		"""Rules of replacement:

		"""
		lhs_args = set(symbol.args)
		if lhs_args != set(free_symbols):
			free_symbols_ = tuple(free_symbols)
			if len(free_symbols) == 1:
				symbol = parse_expr(str(type(symbol)) + str(free_symbols_))
			else:
				if len(lhs_args) > 0:
					raise NameError("Mismatched symbols {} and {}".format(lhs_args, free_symbols))
				else:
					if self.verbose:
						print('type of rhs symbols:', type(free_symbols))
					if type(free_symbols) == set:
						free_symbols_ = sort_symbols(free_symbols)
					if self.verbose:
						print('replacing {} with {}'.format(symbol, str(type(symbol)) + str(free_symbols_)))
					symbol = parse_expr(str(type(symbol)) + str(free_symbols_))

		return symbol

	def validate_function(self, lhs_expr, rhs_expr):
		assert lhs_expr.free_symbols == rhs_expr.free_symbols

	def get_composition(self, lhs_expr, rhs_expr):
		composition = dict()
		for k in list(self.keys()):
			if len(rhs_expr.find(k)) > 0:
				if self.verbose: 
					print('composition detected: found {} in {} = {}'.format(k, lhs_expr, rhs_expr))
				composition[str(k)] = self[k]
		return composition


	def vectorize_function(self, symbol, rhs_expr, composition):
		try:
			func = lambdify(symbol.args, rhs_expr, modules = ['numexpr'])
			if self.verbose:
				print('lambda {} = {} labmdified with numexpr'.format(symbol.args, rhs_expr))
		except:
			func = lambdify(symbol.args, rhs_expr, modules = ['numpy', composition])
			if self.verbose:
				print('lambda {} = {} lambdified with numpy and {}'.format(symbol.args, rhs_expr, composition))
		return func

	def register_signature(self, symbol, units, lhs_expr, rhs_expr):
		self.signatures[str(symbol)] = dict(
				symbol = symbol, 
				units = units, 
				lhs = lhs_expr,
				rhs = rhs_expr,
				# update = getattr(self[lhs_expr],'update', None),
				)

	def register_function(self, func, lhs_symbol, lhs_expr, lhs_units):
		hidden_args = []
		if hasattr(func, 'meta'):
			hidden_args = func.meta.get('hidden_args', [])

		if type(func) is np.vectorize:
			rhs_args = get_function_args(func.pyfunc, hidden_args)
		else:
			rhs_args = get_function_args(func, hidden_args)
		if str(rhs_args[0]) == 'self': # in case function is a class method
			rhs_args.pop(0)
		
		lhs_symbol = self.check_or_replace_symbol(lhs_symbol, rhs_args)
		units = lhs_units
		if hasattr(func, 'meta'):
			if len(lhs_units) > 0:
				if lhs_units != func.meta['units']:
					raise NameError('Units mismatch:{} != {}'.format(lhs_units, func.meta['units']))
			else:
				units = func.meta['units']
			rhs = func.meta.get('equation', func)
		else:
			rhs = func
			try:
				setattr(func, 'meta', dict(units = lhs_units))
			except: # will not work on methods
				pass
		super(Kamodo, self).__setitem__(lhs_symbol, func) # assign key 'f(x)'
		self.register_signature(lhs_symbol, units, lhs_expr, rhs)
		super(Kamodo, self).__setitem__(type(lhs_symbol), self[lhs_symbol]) # assign key 'f'
		self.register_symbol(lhs_symbol) 


	def __setitem__(self, sym_name, input_expr):
		"""Assigns a function or expression to a new symbol, performing unit conversion where necessary

		"""
		if self.verbose:
			print('')
		try:
			symbol, args, units, lhs_expr = self.parse_key(sym_name)
		except KeyError as error:
			if self.verbose:
				print(error)
			found_sym_name = str(error).split('found')[0].strip("'").strip(' ')
			if self.verbose:
				print('replacing {}'.format(found_sym_name))
			self.remove_symbol(found_sym_name)
			symbol, args, units, lhs_expr = self.parse_key(sym_name)

		if hasattr(input_expr, '__call__'):
			self.register_function(input_expr, symbol, lhs_expr, units)
		else:
			rhs_expr = self.parse_value(input_expr, self.symbol_registry)
			units_map = self.get_units_map()
			lhs_units = get_unit(units)
			rhs_units = validate_units(rhs_expr, units_map) # check that rhs units are consistent

			try:
				lhs_units, rhs_units = match_dimensionless_units(lhs_units, rhs_units)
				check_unit_compatibility(rhs_units, lhs_units)
			except:
				print(type(rhs_units))
				print(get_unit(units), validate_units(rhs_expr, units_map))
				raise

			rhs_expr_with_units = get_expr_with_units(rhs_expr, units_map)

			if self.verbose:
				print('rhs_expr with units:', rhs_expr_with_units)

			# convert back to expression without units for lambdify
			if units != '':
				try:
					if self.verbose:
						print('converting to {}'.format(lhs_units))
						for k,v in list(units_map.items()):
							print('\t', k,v, type(k))
					rhs_expr = get_expr_without_units(rhs_expr_with_units, lhs_units, units_map)
					if self.verbose:
						print('rhs_expr without units:', rhs_expr)
				except:
					print('error with units? [{}]'.format(units))
					raise
			else:
				if lhs_units != Dimension(1): # lhs_units were obtained from rhs_units
					units = str(lhs_units)

			rhs_args = rhs_expr.free_symbols

			if self.verbose:
				print('lhs_expr:', lhs_expr, 'units:', lhs_units)
				print('rhs_expr:', rhs_expr, 'units:', rhs_units)
			try:
				symbol = self.check_or_replace_symbol(symbol, rhs_args, rhs_expr)
				self.validate_function(symbol, rhs_expr)

			except:
				if self.verbose:
					print('\n Error in __setitem__', input_expr)
					print(symbol, lhs_expr, rhs_args)
					print('symbol registry:', self.symbol_registry)
					print('signatures:', self.signatures)
				raise
		

			composition = self.get_composition(lhs_expr, rhs_expr)
			func = self.vectorize_function(symbol, rhs_expr, composition)
			self.register_signature(symbol, units, lhs_expr, rhs_expr)
			super(Kamodo, self).__setitem__(symbol, func)
			super(Kamodo, self).__setitem__(type(symbol), self[symbol])
			self.register_symbol(symbol)
			self[symbol].meta = dict(units = units)

	def __getattr__(self, name):
		name_ = symbols(name, cls = Function)
		if name_ in self:
			return self[name_]
		else:
			raise AttributeError("No such symbol: {}".format(name))

	def __delattr__(self, name):
		if name in self:
			del self[name]
		else:
			raise AttributeError("No such field: " + name)

	def get_units_map(self):
		"""Maps from string units to symbolic units"""
		d = dict()
		star = Wild('star')
		for k,v in list(self.signatures.items()):
			unit = get_unit(v['units'])
			d[k] = unit
			d[v['symbol']] = unit

		return d 


	def to_latex(self, keys = None, mode = 'equation'):
		"""Generate list of LaTeX-formated formulas"""
		if keys is None:
			keys = list(self.signatures.keys())
		repr_latex = ""
		for k in keys:
			lhs = self.signatures[k]['symbol']
			rhs = self.signatures[k]['rhs']
			units = self.signatures[k]['units']
			if type(rhs) == str:
				latex_eq = rhs
				# latex_eq = latex(Eq(lhs, parse_latex(rhs)), mode = mode)
			else:
				try:
					latex_eq = latex(Eq(lhs, rhs), mode = mode)
				except:
					lambda_ = symbols('lambda', cls = UndefinedFunction)
					latex_eq = latex(Eq(lhs, lambda_(*lhs.args)), mode = mode)
			if len(units) > 0:
				# latex_eq = latex_eq.replace('=','\\text{' + '[{}]'.format(units) + '} =')
				latex_eq = latex_eq.replace('=','[{}] ='.format(units))
			repr_latex += latex_eq

		return beautify_latex(repr_latex).encode('utf-8').decode()

	def _repr_latex_(self):
		"""Provide notebook rendering of formulas"""
		return self.to_latex()

	def detail(self):
		"""Constructs a pandas dataframe from signatures"""
		return pd.DataFrame(self.signatures).T

	def get_signature(self, name):
		"""Get the signature for the named variable"""
		return self.signatures[str(self.symbol_registry[name])]


	def simulate(self, **kwargs):
		state_funcs = []
		for name, func_key in list(self.symbol_registry.items()):
			func = self[func_key]
			update_var = getattr(func, 'update', None)
			if update_var is not None:
				state_funcs.append((func.update, func))

		return simulate(OrderedDict(state_funcs), **kwargs)

	def evaluate(self, variable, *args, **kwargs):
		if type(self[variable]) is np.vectorize:
			params = get_defaults(self[variable].pyfunc)
		else:
			params = get_defaults(self[variable])

		valid_arguments = valid_args(self[variable], kwargs)
		if params is not None:
			if self.verbose:
				print('default parameters:', list(params.keys()))
			params.update(valid_arguments)
		else:
			if self.verbose:
				print('no default parameters')
			params = valid_arguments
			if self.verbose:
				print('user-supplied args:', list(params.keys()))
		result = eval_func(self[variable], params)
		if isinstance(result, GeneratorType):
			params = concat_solution(result, variable)
		else:
			params.update({variable: result})
		return params

	def solve(self, fprime, interval, y0, 
			dense_output = True, # generate a callable solution
			events = None, # stop when event is triggered
			vectorized = True,):
		from scipy.integrate import solve_ivp
		result = solve_ivp(self[fprime], interval, y0, 
			dense_output = dense_output, 
			events = events, 
			vectorized = vectorized)
		if self.verbose:
			print(result['message'])
		def solution(t = result['t']):
			return result['sol'].__call__(t).T.squeeze()
		return solution


	def figure(self, variable, indexing = 'ij', return_type = False, **kwargs):
		result = self.evaluate(variable, **kwargs)
		signature = self.get_signature(variable)
		units = signature['units']
		if units != '':
			units = '[{}]'.format(units)
		title = self.to_latex([str(self.symbol_registry[variable])], mode = 'inline') #.replace('\\operatorname','')
		title_lhs = title.split(' =')[0]+'$'
		title_short = '{}'.format(variable + units) # something wrong with colorbar latex being vertical
		titles = dict(title = title, title_lhs = title_lhs, title_short = title_short, units = units, variable = variable)
		fig = dict()
		chart_type = None
		traces = []

		hidden_args = []
		if 'hidden_args' in self[variable].meta:
			hidden_args = self[variable].meta['hidden_args']
		arg_arrays = [result[k] for k in result if k not in hidden_args][:-1]

		arg_shapes = get_arg_shapes(*arg_arrays)
		try:
			out_dim, arg_dims = get_plot_key(result[variable].shape, *arg_shapes)
		except:
			print(arg_shapes)
			raise
		try:
			plot_func = plot_dict[out_dim][arg_dims]['func']
		except KeyError:
			print('not supported: out_dim {}, arg_dims {}'.format(out_dim, arg_dims))
			raise
		traces, chart_type, layout = plot_func(result, titles,indexing = indexing, verbose = self.verbose)
		# if len(result) == 2: 
		# 	arg0, val0 = result.items()[0]
		# 	if self.verbose:
		# 		print '1 input: {}-d shape: {}'.format(len(val0.shape), val0.shape)
		# 	if len(val0.shape) == 1:
		# 		lines, chart_type, layout = line_plot(result, titles, self.verbose)
		# 		traces.extend(lines)

		# 	elif len(val0.shape) == 2:
		# 		vectors, chart_type, layout = vector_plot(result, titles, self.verbose)
		# 		traces.extend(vectors)
		# 	else: 
		# 		raise NotImplementedError('shape not supported'.format(val0.shape))
		# elif len(result) == 3: 
		# 	if len(result[variable].shape) == 2:
		# 		contours, chart_type, layout = contour_plot(result, titles, indexing, self.verbose)
		# 		traces.extend(contours)

			
		# elif len(result) == 4:
		# 	result = to_arrays(result) # loses the date-time index
		# 	arg0, val0 = result.items()[0]
		# 	arg1, val1 = result.items()[1]
		# 	arg2, val2 = result.items()[2]
		# 	if self.verbose:
		# 		print '3 inputs: {}{}, {}{}, {}{}'.format(
		# 			arg0, val0.shape, 
		# 			arg1, val1.shape,
		# 			arg2, val2.shape)
		# 	if len(result[variable].shape) == 3:
		# 		planes, chart_type, layout = plane(result, titles, indexing, self.verbose)
		# 		traces.extend(planes)

		# 	elif len(result[variable].shape) == 2:
		# 		if self.verbose:
		# 			print '\t2-d output', result[variable].shape
		# 		surfaces, chart_type, layout = surface(result, titles, self.verbose)
		# 		traces.extend(surfaces)

		# 	elif len(result[variable].shape) == 1:
		# 		if self.verbose:
		# 			print '\tscalar output', result[variable].shape
		# 		if result[variable].shape == val0.shape:
		# 			lines, chart_type, layout = line_plot(result, titles, self.verbose)
		# 			traces.extend(lines)

		# 		elif result[variable].shape[0] == 1:
		# 			surfaces, chart_type, layout = surface(result, titles, self.verbose)
		# 			traces.extend(surfaces)

		# 	elif len(result[variable]) == 1: #scalar or color value?
		# 		surfaces, chart_type, layout = surface(result, titles, self.verbose)
		# 		traces.extend(surfaces)

		# 	else:
		# 		print '{}-d not yet supported'.format(result[variable].shape)
			
		# else:
		# 	print 'not supported yet: {}'.format(type(result[variable]))


		layout.update(
			dict(autosize=False,
				width=700,
				height=400,
				margin=go.layout.Margin(
					# l=30,
					r=30,
					b=32,
					t=40,
					pad=0),
				))


		fig['data'] = traces
		fig['layout'] = layout
		if return_type:
			fig['chart_type'] = chart_type
		return fig

	def plot(self, *variables, **figures):
		for k in variables:
			figures[k] = {}
		if len(figures) == 1:
			variable, kwargs = list(figures.items())[0]
			fig = self.figure(variable, return_type = True, **kwargs)
			if fig['chart_type'] is None:
				raise AttributeError("No chart_type for this trace")
			else:
				if self.verbose:
					print('chart type:', fig['chart_type'])
				return go.Figure(data = fig['data'], layout = fig['layout'])
		else:
			traces = []
			layouts = []
			for variable, kwargs in list(figures.items()):
				fig = self.figure(variable, **kwargs)
				traces.extend(fig['data'])
				layouts.append(fig['layout'])
			return go.Figure(data = traces, layout = layouts[-1])


##------------ Tests Below --------------------

def test_Kamodo_expr():
	a, b, c, x, y, z = symbols('a b c x y z')
	kamodo = Kamodo(a = x**2, verbose = False)
	try:
		assert kamodo.a(3) == 3**2
	except:
		print(kamodo.symbol_registry)
		raise

def test_Kamodo_ambiguous_expr():
	a, b, c, x, y, z = symbols('a b c x y z')
	with pytest.raises(NameError):
		kamodo = Kamodo()
		kamodo[a(b,c)] = x**2 + y**2
	kamodo = Kamodo()
	kamodo['a(x, y)'] = x**2 + y**2
	assert kamodo.a(3,4) == 3**2 + 4**2

def test_Kamodo_callable():
	kamodo = Kamodo(f = lambda x: x**2, verbose = False)
	assert kamodo.f(3) == 9
	kamodo = Kamodo(f = lambda x,y: x**2 + y**3)
	assert kamodo.f(3,4) == 3**2 + 4**3


def test_Kamodo_callable_array():
	kamodo = Kamodo(f = lambda x: x**2)
	assert (kamodo.f(np.linspace(0,1,107)) == np.linspace(0,1,107)**2).all()


def test_Kamodo_str():
	kamodo = Kamodo('f(a,x,b) = a*x+b')
	try:
		assert kamodo.f(3,4,5) == 3*4+5
	except:
		print(kamodo.signatures)
		print(list(kamodo.keys()))
		raise

def test_Kamodo_latex():
	kamodo = Kamodo('$f(a,x,b) = a^x+b $')
	assert kamodo.f(3,4,5) == 3**4+5

def test_Kamodo_get():
	kamodo = Kamodo('$f(a,x,b) = a^x+b $')
	try:
		assert kamodo['f'](3,4,5) == 3**4+5
	except:
		print(kamodo.signatures)
		print(list(kamodo.keys()))
		raise


def test_Kamodo_set():
	kamodo = Kamodo()
	kamodo['f(a,x,b)'] = '$a^x+b$'
	assert kamodo.f(2,3,4) == 2**3+4


def test_Kamodo_mismatched_symbols():
	with pytest.raises(NameError):
		kamodo = Kamodo('$f(a,b) = a + b + c$', verbose = False)
		assert 'f' not in kamodo

def test_Kamodo_assign_by_expression():
	kamodo = Kamodo()
	f,x,y = symbols('f x y')
	kamodo[f(x,y)] = x**2 + y**2
	assert kamodo['f'](3,4) == 3**2 + 4**2
	assert kamodo.f(3,4) == 3**2 + 4**2 

def test_Kamodo_assign_by_callable():
	kamodo = Kamodo(f = lambda x: x**2, verbose = False)
	assert kamodo['f'](3) == 3**2
	assert kamodo.f(3) == 3**2	
	Kamodo(f = lambda x,y: x+y)
	def rho(x,y): 
		return x+y
	kamodo['g'] = rho
	assert kamodo.g(3,4) == 3 + 4

def test_Kamodo_composition():
	f, g = symbols('f g', cls = UndefinedFunction)
	x = Symbol('x')
	kamodo = Kamodo(f = "$x^2$")
	kamodo['g(x)'] = "$f(x) + x^2$"
	kamodo['h(x)'] = g(x)**2 + x**2

	assert kamodo.f(3) == 3**2
	assert kamodo.g(3) == 3**2 + 3**2
	assert kamodo.h(3) == (3**2 + 3**2)**2 + 3**2

def test_Kamodo_reassignment():
	a,b,c, x,y,z, r = symbols('a b c x y z r')
	kamodo = Kamodo('f(x) = 3*x+5')
	kamodo[r(x,y)] = x + y
	kamodo[r(x,y)] = "3*x + y"

def test_multivariate_composition():
	kamodo = Kamodo(f = 'x**2', g = lambda y: y**3)
	kamodo['h(x,y)'] = 'f(x) + g(y)'
	kamodo['i(x,y)'] = 'x + f(h(x,y))'
	assert kamodo.h(3,4) == 3**2 + 4**3
	assert kamodo.i(3,4) == 3 + (3**2 + 4**3)**2


def test_special_numbers():
	kamodo = Kamodo()
	kamodo['f'] = "${}^x$".format(np.e)
	assert np.isclose(kamodo.f(1), np.e)

def test_function_registry():
	kamodo = Kamodo("f(x) = x**2", "g(y) = y**3")
	kamodo['r(x,y)'] = "(x**2 + y**2)**(1/2)"
	kamodo['h(x,y)'] = 'f + g + r'
	assert 'h(x, y)' in kamodo.signatures   

def test_symbol_key():
	f = symbols('f', cls = UndefinedFunction)
	x = Symbol('x')
	f_ = f(x)
	kamodo = Kamodo()
	kamodo[f_] = x**2
	assert kamodo.f(3) == 9

def test_compact_variable_syntax():
	f = symbols('f', cls = UndefinedFunction)
	x = symbols('x') 
	kamodo = Kamodo(f = 'x**2')# f also registers f(x)
	kamodo['g(x)'] = 'f + x'
	assert kamodo.g(3) == 3**2 + 3

def test_unit_registry():
	def rho(x,y):
		return x*y
	def v(x,y):
		return x+y
	v.meta = dict(units = 'km/s')
	kamodo = Kamodo(v = v, verbose = False)
	kamodo['rho[kg/cc]'] = rho
	try:
		assert kamodo.rho.meta['units'] == 'kg/cc'
		assert kamodo.v.meta['units'] == 'km/s'
	except:
		print('\n',pd.DataFrame(kamodo.signatures))
		raise

	with pytest.raises(NameError):
		kamodo['p[kg]'] = v
	assert 'p' not in kamodo

def test_to_latex():
	kamodo = Kamodo(f = 'x**2')
	assert kamodo.to_latex() == '\\begin{equation}f{\\left (x \\right )} = x^{2}\\end{equation}'
	kamodo = Kamodo(g = 'x')
	assert kamodo.to_latex() == '\\begin{equation}g{\\left (x \\right )} = x\\end{equation}'


def test_expr_conversion():
	kamodo = Kamodo('$a[km] = x$')
	kamodo.a

def test_validate_units():
	f,x = symbols('f x')
	with pytest.raises(ValueError):
		validate_units(Eq(f,x), dict(f = get_unit('kg'), x = get_unit('m')))
	validate_units(Eq(f,x), dict(f = get_unit('kg'), x = get_unit('g')))
	
	lhs_units = validate_units(f(x), dict(f = get_unit('kg'), x = get_unit('m')))
	print(lhs_units)

def test_unit_conversion():
	kamodo = Kamodo('$a(x)[km/s] = x$',
						'$b(y)[m/s]  = y$',)
	kamodo['c(x,y)[m/s]'] = '$a + b$'
	assert kamodo.c(1,2) == 1002 


def test_get_unit():
	assert get_unit('kg/m^3') == get_unit('kg/m**3')

def test_unit_conversion_syntax():
	kamodo = Kamodo('rho[kg/m^3] = x', verbose = False)
	with pytest.raises(NameError):
		kamodo['d[kg]'] = 'rho'
		print(kamodo.detail())


def test_unit_composition():
	kamodo = Kamodo('m[kg] = x', verbose = False)
	kamodo['v[km/s]'] = 'y'
	kamodo['p'] = 'm*v'
	assert get_unit(kamodo.signatures['p(x, y)']['units']) == get_unit('kg*km/s')

def test_unit_function_composition():
	kamodo = Kamodo('X[m] = x', verbose = True)
	@kamodofy(units = 'km/s')
	def v(x):
		return y

	kamodo['v'] = v
	kamodo['speed'] = 'v(X)'
	assert kamodo.speed.meta['units'] == 'kilometer/second'



def test_method_args():
	class TestModel(Kamodo):
		def __init__(self, *args, **kwargs):
			super(TestModel, self).__init__(*args, **kwargs)
			self['rho'] = self.density
				
		def density(self, alt, lat, lon):
			return alt + lat + lon
		
		density.meta = dict(units = '1/cm^3')

	test = TestModel(verbose = False)
	assert str(list(test.keys())[0].args[0]) != 'self'

	test['r(alt, lat, lon)[1/m^3]'] = 'rho'
	try: # check that args are in the correct order
		assert list(test.signatures.values())[0]['symbol'].args == list(test.signatures.values())[1]['symbol'].args
	except:
		print('\n', test.detail())
		raise



def test_komodofy_decorator():
	@kamodofy(units = 'kg/cm^3')
	def my_density(x,y,z):
		return x+y+z

	assert my_density.meta['units'] == 'kg/cm^3'
	assert my_density(3,4,5) == 12

def test_komodofy_register():
	@kamodofy(units = 'kg/cm^3')
	def my_density(x,y,z):
		return x+y+z

	kamodo = Kamodo(rho = my_density)
	assert kamodo.rho.meta['units'] == 'kg/cm^3'

def test_komodofy_method():
	class TestModel(Kamodo):
		def __init__(self, *args, **kwargs):
			super(TestModel, self).__init__(*args, **kwargs)
			self['rho'] = self.density
		
		@kamodofy(units = '1/cm^3')        
		def density(self, alt, lat, lon):
			return alt + lat + lon

	test = TestModel(verbose = False)
	assert test.density.meta['units'] == '1/cm^3'



# 	try:
# 		assert len(kamodo.detail()) == 1
# 	except:
# 		print kamodo.symbol_registry
# 		print kamodo.detail()
# 		raise

# # def test_remove_symbol():
# # 	kamodo = Kamodo(f = 'x')
# # 	kamodo['g'] = '2*f'
# # 	kamodo.remove_symbol('f')
# # 	try:
# # 		assert len(kamodo.symbol_registry) == 1
# # 		assert len(kamodo) == 2
# # 	except:
# # 		print '\n',kamodo.detail()
# # 		raise

# def test_function_arg_ordering():
# 	def x(r,theta,phi):
# 	    '''converts from spherical to cartesian'''
# 	    return r*np.sin(theta)*np.cos(phi)

# 	kamodo = Kamodo(x = x)
# 	rhs_args = get_function_args(kamodo.x)
# 	lhs_args = kamodo.signatures.values()[0]['symbol'].args

# 	for i in range(3):
# 	    assert lhs_args[i] == rhs_args[i]

# def test_function_change_of_variables():
# 	def rho(x):
# 		return x**2

# 	kamodo = Kamodo(x = 'y')
# 	kamodo['rho'] = rho
# 	kamodo['rho_perp'] = 'rho(x)'

# 	assert str(get_function_args(kamodo.rho_perp)[0]) == 'y'


def test_kamodofy_update_attribute():
	@kamodofy(units = 'cm', update = 'y')
	def f(x):
		return x

	kamodo = Kamodo(f = f)
	assert f.update == 'y'


def test_kamodo_coupling():
	@kamodofy(units = 'cm', update = 'y')
	def y_iplus1(y, x):
		return y + x

	@kamodofy(units = 'm')
	def x_iplus1(x, y):
		return x - y


	kamodo = Kamodo()
	kamodo['x_iplus1'] = x_iplus1
	kamodo['y_iplus1'] = y_iplus1

	kamodo.x_iplus1.update = 'x'
	assert kamodo.x_iplus1.update == 'x'
	assert kamodo.y_iplus1.update == 'y'

	simulation = kamodo.simulate(y = 1, x = 0, steps = 1)

	for state in simulation:
		pass

	assert state['x'] == -1
	assert state['y'] == 0


def test_units():
	nT = get_unit('nT')
	assert str(nT) == 'nanotesla'
	assert str(nT.abbrev) == 'nT'

	R_E = get_unit('R_E')
	assert str(R_E) == 'earth radii'
	assert str(R_E.abbrev) == 'R_E'

def test_default_composition():
	## Need to wrap function defaults
	def create_wrapped(args, expr):
		func = lambdify(args, expr)
		@functools.wraps(func)
		def wrapped(*_args, **kwargs):
			# Write the logic here to parse _args and kwargs for the arguments as you want them
			return func(*actual_args)
		return wrapped

def test_vectorize():
	@np.vectorize
	@kamodofy(units = 'kg')
	def f(x, y):
		return x + y

	kamodo = Kamodo(f = f)
	kamodo.f([3,4,5],5)
	


def test_redefine_variable():
	kamodo = Kamodo(rho = 'x + y')
	kamodo['rho'] = 'a + b'
	kamodo['rho(a,b)'] = 'a*b'


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = 'kamodo',
    version = '0.0.2',
    author = 'Asher Pembroke',
    author_email = 'apembroke@gmail.com',
    description = 'A functional api for scientific data',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://ccmc.gsfc.nasa.gov/Kamodo/',
    packages = setuptools.find_packages(),
    classifiers = [
		"Programming Language :: Python :: 3.7",
	    "Operating System :: OS Independent",
	    "License :: OSI Approved",
    ],
    install_requires = [
		'numpy',
		'scipy',
		'sympy',
		'pandas',
		'plotly',
		'pytest',
        'click',
    	],
    license='NASA OPEN SOURCE AGREEMENT VERSION 1.3',

)

